************************************************************************
*
* 8 bit PC bmp converter & loader
* this loads an 8 bit (256 colour) PC Bitmap directly into the screen
* memory of a channel, provided this is in mode 32 or 33
*
* syntax:
*
*	BMP8LOAD [#chan%,] file_name$
*		Chan% defaults to 1
*
* the image will be clipped if necesary
*
* v. 1.00 2002 Mar 16 17:15:36 initial version
* v. 1.01 2002 Mar 27 07:39:51 can also convert for Q60/Q40
* v. 1.02 2002 Apr 12 18:57:02 uses external op, test for Q60 is made according to screen mode
* (c) W. Lenerz 2002 - 2003
*************************************************************************


	section cvt

	xdef		wl_bmp8load
	include 	DEV8_keys_con
	include 	dev8_keys_qdos_io
	include 	dev8_keys_qdos_sms
	include 	dev8_keys_err


; some macros

xjsr	macro	name
	xref	[name]			; - makes external jump
	jsr	[name]
	endm

dtrap	macro	nbr,name
	trap	#[nbr]			; trap with error check
	tst.l	d0
	bne	[name]
	endm

; in ql screen (mode 32)  the colour bits are : gggbbbbb rrrrrggg <- qpc
; in ql screen (mode 33)  the colour bits are : gggggrrr rrbbbbbw <- q60
;	where w = intensity. set to 1 if at least two of the 6th bits in the
;	pc screen) for rgb are 1
; in pc colour bitmap, the bits are bgr, and the lines are bottom down!
; the palette is BGR with 1 space padding at end
;


space	equ	$400
fsize	equ	2			; total file size
offset	equ	$a			; offset to bitmap data
hsize	equ	$e			; header size
width	equ	$12			; x size of file is here (pixels)
height	equ	$16			; y size of file is here (pixels)
planes	equ	$1a			; nbr of planes
type	equ	$1c			; type of bitmap, must be 24 for 24 bit image
compress equ	$1e			; type of compression
size	equ	$22			; size of the bitmap data in bytes. this number must be rounded to the next 4 byte boundary
nbcol	equ	$2e			; nbr of cols used
impcol	equ	$32			; nbr of important colours (?!)
hlen	equ	$36			; header length, here starts the data (or palette)
longw	equ	%11111000111111001111100000000000
wqx0	equ	%00000100000001000000010000000000



wl_bmp8load
; first get the input filename, open file & load image into mem
	moveq	#-1,d1
	moveq	#-1,d2
	moveq	#sms.dmod,d0
	trap	#1			; get screen mode
	move.w	d1,d5			; keep it
	sub.w	#32,d1			; it must be at least mode 32
	blt	no_good 		; but it isn't->...
	xjsr	chan_num		; get channel id, default to #1
	bne	bmp_out 		; oops! ->...

	moveq	#0,d1
	moveq	#0,d2
	moveq	#iop.slnk,d0		; pretend to set bytes in linkage block
	trap	#3
	tst.l	d0
	bne	r_out
	clr.l	d6
	move.w	pt_scinc(a1),d7 	; line length in bytes
	move.w	pt_xscrs(a1),d1
	divs.w	d1,d7			; nbr of bytes/pixel
	tst.w	d7
	beq	no_good 		; there must be at least 1 byte/pixel!
	move.l	a0,a2			; keep
	move.l	#space,d0		; get mem & string from basic into mem
	xjsr	cpy_str 		; pointed then by a0
	bne	bmp_out 		; oops! ->...
	move.l	a0,a4			; keep pointer to filename space
	moveq	#1,d3			; shared access
	moveq	#1,d0
	moveq	#-1,d1
	dtrap	2,bmp_out2		; open file
	move.l	a4,a1
	move.l	#$100,d2
	moveq	#iof.rhdr,d0
	moveq	#-1,d3
	dtrap	3,bmp_out2		; read file header into (a1)
	move.l	(a4),d4 		; length of file
	exg	a0,a4
	xjsr	gu_rchp 		; release mem for filename
	move.l	d4,d0			; length of file
	add.l	#space,d0		; + some space for myself
	xjsr	gu_achp 		; get mem for the file
	bne	bmp_out4		; oops ->...
	exg	a0,a4			; a0 = channel id, a4 = memptr
	lea	space(a4),a1		; point to space for file
	move.l	d4,d2
	moveq	#iof.load,d0		; (scatter) load file
	dtrap	3,bmp_out3
	moveq	#2,d0
	trap	#2			; close input file now

; file is loaded into space(a4) and closed

get_size
	moveq	#err.nimp,d0		; error nimp if wrong file type
	move.w	width+space(a4),d6	; image x size (nbr pix)in intel format
	beq	bmp_out 		; what, 0 sized image???? ->...
	rol.w	#8,d6			; now in comprehensible format
	swap	d6
	move.w	height+space(a4),d6	; image y size - nbr of lines
	beq	bmp_out 		; what, 0 sized image???? ->...
	rol.w	#8,d6			; now d6 = pc xsize | ysize
	move.l	a4,a5			; point to place for my palette

	cmp.b	#$8,type+space(a4)	; file type 8 bit map??
	bne	bmp24			; ....no ->...

; now make my palette map
; d5 = screen mode (33/32)
; d6 = PC bitmap x size | y size in nbr of pixels
; a1 points to the pc bmp palette
; a4 = my meory area
; a5 points to what will be my ql screen palette map

allmods move.w	space+nbcol(a4),d3	; nbr of colours for this bitmap
	bne.s	no_0
	move.w	#255,d3 		; simulate 256 colours palette
	bra.s	com
no_0	rol.w	#8,d3
	subq.w	#1,d3
com	lea	space+hlen(a4),a3	; point to palette
	cmp.b	#32,d5			; mode 32?
	bne.s	bord33			; no, so must be 33->...

;					  result wished:   gggbbbbbrrrrrggg
bord32	move.l	(a3)+,d0		; bbbbbbbbgggggggg rrrrrrrrxxxxxxxx
	rol.w	#5,d0			; bbbbbbbbgggggggg rrrxxxxxxxxrrrrr
	swap	d0			; rrrxxxxxxxxrrrrr bbbbbbbbgggggggg
	lsr.b	#2,d0			; rrrxxxxxxxxrrrrr bbbbbbbb00gggggg
	ror.w	#6,d0			; rrrxxxxxxxxrrrrr ggggggbbbbbbbb00
	ror.l	#5,d0			; bbb00rrrxxxxxxxx rrrrrggggggbbbbb
	ror.w	#8,d0			; bbb00rrrxxxxxxxx gggbbbbbrrrrrggg
	move.w	d0,(a5)+		;
	dbf	d3,bord32
	bra.s	pal_end
;					  result wished:   rrrrrgggggbbbbbw
bord33	move.l	(a3)+,d1
	move.l	d1,d0			; bbbbbbbbgggggggg rrrrrrrrxxxxxxxx
	rol.w	#5,d0			; bbbbbbbbgggggggg rrrxxxxxxxxrrrrr
	swap	d0			; rrrxxxxxxxxrrrrr bbbbbbbbgggggggg
	lsr.b	#3,d0			; rrrxxxxxxxxrrrrr bbbbbbbb000ggggg
	ror.w	#5,d0			; rrrxxxxxxxxrrrrr gggggbbbbbbbb000
	ror.l	#5,d0			; bbb00rrrxxxxxxxx rrrrrgggggbbbbbb
	bclr	#0,d0			; bbb00rrrxxxxxxxx rrrrrgggggbbbbb0
					; 00000100 00000100 00000100 00000000
; got correct bits, now make white bit	:      26	18	 10
	andi.l	#wqx0,d1
	rol.l	#6,d1			; accumulate the 3 bits
	move.b	d1,d2			; first bit (B)
	rol.l	#8,d1			; second bit G)
	add.b	d1,d2
	rol.l	#8,d1
	add.b	d1,d2			; third bit (R)
	subq.b	#2,d2			; must be at least 2 to set w bit
	bmi.s	put			; but it isn't
	bset	#0,d0
put	move.w	d0,(a5)+		;
	dbf	d3,bord33

; ok the palette is transformed

pal_end move.l	a2,a0			; point to channel id given as parameter
	lea	xtop8,a2
commn
	sub.l	#16,a7
	move.l	a7,a1
	move.b	d5,d2			; screen mode
	move.l	a4,(a1) 		; point to my mem
	move.w	d7,4(a1)		; nbr of bytes/pixel
	move.l	d6,6(a1)		; pc image screen sizes
	moveq	#iow.xtop,d0
	trap	#3			; do external op
bmp_ok
	add.l	#16,a7
	clr.l	d0
bmp_out2
	move.l	a4,a0
	xjsr	gu_rchp
bmp_out rts

bmp_out3
	moveq	#2,d0
	trap	#2
	bra.s	bmp_out2
bmp_out4
	move.l	d0,d6
	moveq	#2,d0
	trap	#2
	move.l	d6,d0
	rts

no_good moveq	#err.ipar,d0
r_out	rts

bmp24	move.l	a2,a0			; point to channel ID
	lea	xtop24,a2		; external routine
	bra.s	commn			; do it




; this is the external op for 8 bit colours
; it actually copies bytes into screen mem
xtopreg reg	d2-d7/a0/a2-a5
xtop8	movem.l xtopreg,-(a7)		; keep the regs

	move.l	(a1),a4 		; pointer to my memory
	move.l	sd_scrb(a0),a3		; point to screen base
	move.w	4(a1),d6		; nbr bytes/pixel
	move.l	6(a1),d7		; pc image size
	clr.l	d4
	move.w	sd_linel(a0),d4 	; 1 screen line: length in bytes
	move.b	d2,d5
	move.w	sd_xmin(a0),d0		; channel xorig
	mulu	d6,d0
	add.l	d0,a3
	move.w	sd_ymin(a0),d0		; channel y orig
	mulu	d4,d0
	add.l	d0,a3			; a3 points to orig of my channel in screen mem
	move.l	sd_xsize(a0),d0 	; channel x,y size

; now possibly clip image
	move.w	d7,d2			; image y size
	cmp.w	d2,d0
	bge.s	y_ok
	move.w	d0,d2
y_ok	move.l	d7,d1
	swap	d1			; image x size
	move.w	d1,d3			; image x size (stays in bits)
	clr.l	d5
	move.w	d3,d5
	divs	#4,d5
	swap	d5
	tst.w	d5
	beq.s	none
	neg.w	d5
	addq.w	#4,d5
	add.w	d5,d3			; make sure this is a multiple of 4
none	swap	d0
	cmp.w	d1,d0
	bge.s	x_ok
	move.w	d0,d1
x_ok	subq.w	#1,d1
	subq.w	#1,d2

; a0 = scr driver defn block for this channel
; a1 = palette
; a3 = screen
; a4 = data (palette + bmp)
; a5 = ?
; a6 = sys vars

	move.l	a4,a1			; my palette
	lea	space(a4),a2		; point to file data bits - first line
	move.l	offset(a2),d0		; offset
	rol.w	#8,d0
	swap	d0
	rol.w	#8,d0
	add.l	d0,a2			; point to data bits now
; the data bits in a pc bitmap start at the bottom (!)
	move.w	d7,d5			; image y size
	subq.w	#1,d5
	mulu	d3,d5
	add.l	d5,a2			; point to last row
lp1	move.l	a2,a4			; point data bits
	move.l	a3,a5			; point screen
	move.w	d1,d0
lp2	clr.w	d5
	move.b	(a4)+,d5		; one data bits
	lsl.w	#1,d5			; it is a word sized palette
	move.w	(a1,d5.w),(a5)+ 	; put corresponding word in screen
	dbf	d0,lp2			; for all pixels in a row
	add.l	d4,a3			; point to new screen row
	sub.w	d3,a2			; point to new data bits
	dbf	d2,lp1
	movem.l (a7)+,xtopreg
	clr.l	d0
	rts

*--------------------------------------------
; this is the external op for 24 bit colours
; it actually copies bytes into screen mem
; d2 = screen mode
; a1 pointer to some memory for myself
*--------------------------------------------
xtop24
	movem.l xtopreg,-(a7)		; keep the regs
	move.l	(a1),a4 		; pointer to my memory
	move.l	sd_scrb(a0),a3		; point to screen base
	move.w	4(a1),d6		; nbr bytes/pixel
	move.l	6(a1),d7		; pc image size
	clr.l	d5
	move.w	sd_linel(a0),d5 	; 1 screen line: length in bytes

	move.w	sd_xmin(a0),d0		; channel xorig
	mulu	d6,d0
	add.l	d0,a3
	move.w	sd_ymin(a0),d0		; channel y orig
	mulu	d5,d0
	add.l	d0,a3			; a3 points to orig of my channel in screen mem
	move.l	sd_xsize(a0),d0 	; channel x,y size
	move.b	d2,d4			; screen mode
; now possibly clip image
	move.w	d7,d2			; PC bitmap image y size
	cmp.w	d2,d0			; bigger than channel size?
	bge.s	y_ok2			;  ... no
	move.w	d0,d2			;  ... yes, use channel x size
y_ok2	move.l	d7,d1			; PC bitmap xsize | ysize
	swap	d1			; PC bitmap image x size
	move.w	d1,d3			; image x size (stays in bytes)
	mulu	#3,d3			; 3 bytes per pixel
	clr.l	d5
	move.w	d3,d5
	divs	#4,d5
	swap	d5
	tst.w	d5
	beq.s	none2
	neg.w	d5
	addq.w	#4,d5
	add.w	d5,d3			; make sure this is a multiple of 4
none2	swap	d0			;
	cmp.w	d1,d0			; bitmap x size > channel x size ?
	bge.s	x_ok2			;  ... no
	move.w	d0,d1			;  ... yes, use channel x size
x_ok2	subq.w	#1,d1			; prepare for dbf
	subq.w	#1,d2			;
; d1 = nbr of lines to treat -1
; d2 = nbr of pixels per line to treat -1
; d3 = nbr of bytes per line in bitmap
; a0 = scr driver defn block for this channel
; a3 = screen
; a4 = data (palette + bmp)
; a5 = ?
; a6 = sys vars

	lea	space+hlen(a4),a2	; point to file data bits - first line
; the data bits in a pc bitmap start at the bottom (!)
	move.w	d7,d5			; image y size
	subq.w	#1,d5
	mulu	d3,d5
	add.l	d5,a2			; point to last row
	cmp.b	#32,d4			; mode 32?
	bne.s	md33			; no....

; here mode 32				   gggbbbbb rrrrrgggg
md32	clr.l	d4
	move.w	sd_linel(a0),d4 	; 1 screen line: length in bytes
oloop32 				; outer loop - treats each line
	move.l	a2,a4			; point data bits
	move.l	a3,a5			; point screen
	move.w	d1,d7			; do it for d7 pixels per line
; this is the inner loop:  covers each pixel in a line (row)
iloop32
	move.b	(a4)+,d0		; blue
	lsr.b	#3,d0			; get rid of the 3 lsbs
	move.b	(a4)+,d5		; green - upper bits go with red, lower with blue
	rol.b	#3,d5			; wrap it around
	move.b	d5,d6			; keep
	and.b	#%11100000,d6		; isolate 3 green bits
	or.b	d6,d0			; and make gggbbbbb
	lsl.w	#8,d0			; move into upper byte
	and.b	#%00000111,d5		; upper 3 bits for green
	move.b	(a4)+,d0		; red
	and.b	#%11111000,d0		; mask out
	or.b	d5,d0			; set now
	move.w	d0,(a5)+
	dbf	d7,iloop32

	add.l	d4,a3			; point to new screen row
	sub.w	d3,a2			; point to new data bits
	dbf	d2,oloop32
	movem.l (a7)+,xtopreg
	clr.l	d0
	rts
; here mode 33

md33	clr.l	d4
	move.w	sd_linel(a0),d4 	; 1 screen line: length in bytes
oloop33
	move.l	a2,a4			; point data bits
	move.l	a3,a5			; point screen
	move.w	d1,d7			; do it for d0 pixels per line

; this is the inner loop:  covers each pixel in a line (row) -	bgr
; due to the need to check the 2nd bit of each colour word, this is quite a
; bit longer than the mode32 loop!
iloop33
	clr.w	d6
	move.b	1(a4),d0		; green 	       xxxxxxxx gggggggg
	andi.b	#%11111100,d0		; get rid of the 2 lowest bits
	bclr	#2,d0			; also test & clear bit 2
	beq.s	nobit7			; bit wasn't set, anyway
	moveq	#1,d6			; show that one bit was set
nobit7	lsl.w	#8,d0			; move into upper byte: gggggxxx xxxxxxxx
	move.b	2(a4),d5		; red
	andi.w	#%11111100,d5		; get rid of 2 lowest bits (and upper byte)
	bclr	#2,d5			; also test & clear bit 2
	beq.s	nobit8
	addq.b	#1,d2
nobit8	lsl.w	#3,d5
	or.w	d5,d0			;
	move.b	(a4),d5 		; blue
	bclr	#2,d5			; test & clear
	beq.s	nobit9
	addq.b	#1,d6			; show bit was set
nobit9	lsr.b	#2,d5
	or.b	d5,d0
	subq.b	#2,d6
	bmi.s	nobit10
	bset	#0,d0
nobit10 move.w	d0,(a5)+
	addq.l	#3,a4			; point next bgr
	dbf	d7,iloop33		; treat 1 line
	add.l	d4,a3			; point to new screen row
	sub.w	d3,a2			; point to new data bits
	dbf	d2,oloop33
	movem.l (a7)+,xtopreg
	clr.l	d0
	rts
	end
