* bmp to QL sprite converter
* (c) W. Lenerz 2003

* syntax:
* result = WL_BMP2SPRT64  (file_in$,file_out$[,x%,y%][,alpha%])
*
* converts a PC 8 or 24 bits bitmap into a 24 bits QL sprite
* file_in$ is the name of the bitmap file
* file_out$ is the (possible!) file for the resulting QL sprite
* if x% and y% are given, the colour of that spot will be made transparent
* if alpha% is given, use alpha channel for sprite
* the function returns
*		nagtive rror code if error
*		0 if all ok and file_out$ given
*		> 0 if all OK and file_out$ was "": the function returns the
*		    address of the sprite
*
* 2004 Jan 17 08:52:10
* v. 1.02	OK if alpha mask & tot nbr of pixels not divisable by 4
* v. 1.01	handles 8 bit images correctly (lines padded to long word limit)
* v. 1.00	initial version

	section cvt

	xdef	WL_bmp2sprt64
	xref	rt_fp
	include dev8_keys_qlv
	include DEV8_keys_qdos_pt


; some macros

xjsr	macro	name
	xref	[name]			; - makes external jump
	jsr	[name]
	endm

dtrap	macro	nbr,name
	trap	#[nbr]			; trap with error check
	tst.l	d0
	bne	[name]
	endm

; in pc 24 colour bitmap, the bits are bgr, and the lines are bottom down!
;

fname	equ	0
palet	equ	fname+256
space	equ	4*256+palet

width	equ	$12			; x size of file is here (pixels)
height	equ	$16			; y size of file is here (pixels)
type	equ	$1c			; type of bitmap, must be 24 for 24 bit image, 8 for 8 bit
nbcol	equ	$2e
hlen	equ	$36


err_rts moveq	#-15,d0
out	rts

; first get the input filename, open file & load image into mem
WL_bmp2sprt64
	moveq	#-1,d4			; preset no optional params
	move.l	a5,d0
	sub.l	a3,d0
	lsr.w	#3,d0			; nbr of parameters passed
	subq.w	#2,d0			; are there any optional parameters?
	beq.s	no_opts 		;   ... no
	lea	16(a3),a3		; cut out first 2 params
	move.w	sb.gtint,a2
	jsr	(a2)			; get integers
	bne.s	out			; oops
	move.w	d3,d0
	lsl.w	#1,d0
	add.l	d0,$58(a6)		; reset maths stack
	subq.b	#1,d3			; only one optional param?
	beq.s	is_alph
	move.l	(a6,a1.l),d4		; params (colour spot)
	addq.l	#4,a1			; point possible next parameter
	subq.w	#1,d3			; 2 or 3 optional params?
	beq.s	no_alph 		; 2
is_alph move.w	(a6,a1.l),d7		; possible alpha channel
	beq.s	no_alph 		; there is none
	move.w	#pto.alph,d7		 ; show alpha flag
no_alph move.l	a3,a5
	lea	-16(a3),a3

no_opts move.l	#space,d0		; get string from basic, into space
	xjsr	cpy_str 		; pointed then by a0
	bne.s	out			; oops!
	move.l	a0,a4			; keep pointer to filename space
	moveq	#1,d3			; file open - shared access key
	moveq	#1,d0
	moveq	#-1,d1
	dtrap	2,bmp_out2		; open file
	move.l	a4,a1			; pointer to some memory
	move.l	#space,d2		; max space for header
	moveq	#$47,d0
	moveq	#-1,d3
	dtrap	3,bmp_out2		; read file header
	move.l	(a4),d5 		; length of file
	exg	a0,a4
	xjsr	gu_rchp 		; release mem for filename
	move.l	d5,d0			; length of file
	add.l	#space,d0		; + some space for myself
	xjsr	gu_achp 		; get mem for the file
	bne	out2			; oops->... (file stays open!!!)

	exg	a0,a4			; a0 = channel id, a4 = memptr
	lea	space(a4),a1		; point to space for file
	move.l	d5,d2
	moveq	#$48,d0 		; greek goddess (scatter load file)
	dtrap	3,bmp_out3
	moveq	#2,d0
	trap	#2			; close input file now

	move.l	a4,a0			; keep mem pointer
	move.l	#space-12,d0		; max length of output filename
	move.l	$58(a6),a1		; point maths stack
	addq.l	#8,a3
	xjsr	get_str 		; get string now into(a4) (result filename)
	beq.s	get_size		; ok, got it
	cmp.l	#-7,d0			; 0 length string?
	bne	bmp_out2		 ; oops!->...

get_size
	clr.l	d1
	clr.l	d2
	clr.l	d5
	clr.l	d6
	moveq	#-15,d0 		; error bp if wrong file type
	move.w	d7,d3
	swap	d3			; d3 = alpha flag
	move.b	type+space(a4),d3	; file type
	subq.b	#$8,d3			; 8 bit bitmap?
	beq.s	bmp_ok			;   ...yes
	sub.b	#16,d3			; 24 bit ?
	bne	bmp_out2		;   ...no, can't handle it, then
	move.w	#-1,d3			; show 24 bit
bmp_ok	move.w	width+space(a4),d7	; image x size (nbr pix)in intel format
	beq	bmp_out2		; what, 0 sized image???? ->...
	rol.w	#8,d7			; now in comprehensible format
	move.w	d7,d1			; keep
	swap	d7
	move.w	height+space(a4),d7	; image y size - nbr of lines
	beq	bmp_out2		; what, 0 sized image???? ->...
	rol.w	#8,d7			; now d7 = pc xsize | ysize
	move.w	d7,d0			; d0 = pc ysize
	cmp.w	d0,d4			; is "origin" within sprite?
	ble.s	gt_size 		; yes
	bra.s	bad_org 		; no
	swap	d4
	cmp.w	d1,d4
	swap	d4
	ble.s	gt_size
bad_org moveq	#-1,d4			; no origin

gt_size mulu	d1,d0			; xsize * y size... = nbr pixels
	move.l	d0,d5			; keep
	lsl.l	#3,d0			; *8 = nbr of bytes needed for ql sprite (pattern + mask)
	add.l	#pto_blk+8,d0		; plus header
	move.l	d0,d6			; keep
	addq.l	#4,d0			; space for rechp line
	xjsr	gu_achp 		; get mem
	bne	bmp_out2		; ooops ->...
	lea	$e0(a6),a1		; pointer to our link
	move.w	$d2,a2			; add this mem to my basic heap
	jsr	(a2)			; link this into heap
	addq.l	#4,a0			; cut out linkage area, for rechp
	move.l	a0,a2			; keep
	move.l	a2,a5
	lea	pto_blk+4(a2),a3	; start of actual data - pattern
	move.w	#$240,(a2)+		; pto_form - 24 bits sprite
	swap	d3
	move.w	d3,(a2)+		; possible alpha channel
	swap	D3
	move.l	d7,(a2)+		; size
	addq.l	#4,a2			; jump over offsets
	move.l	a3,d0
	sub.l	a2,d0
	move.l	d0,(a2)+		; pointer to colour pattern
			
; A2 = pointer into sprite header
; A4 = pointer to mem area, PC bmp lays at space(a4)
; A5 = pointer to space for sprite
; d3 = .H = alpha flag .L if -1 if 24 bitmap, 0 if 8 bitmap
; d4 = -1 if no optional params, else x,y spot
; d5 = number of pixels in sprite
; d6 = size of sprite + heabder in bytes
; d7 = sprite ysize | sprite xsize
	clr.l	d0
	clr.l	d1
	move.w	d7,d0			; nbr of lines in bitmap
	subq.w	#1,d0			; one less now
	swap	d7
	move.w	d7,d1			; nbr of pixels per lines in bitmap
	tst.b	d3
	beq.s	o8bit
	mulu	#3,d1			; nbr of bytes per line
o8bit	move.l	d1,d2
	divs	#4,d2
	swap	d2
	tst.w	d2			; any padding needed?
	beq.s	no_pad			; ... no
	neg.w	d2
	addq.w	#4,d2
	add.w	d2,d1			; padding at end

no_pad	move.l	d0,d2
	mulu	d1,d2
	subq.w	#1,d7			; sprite x size -1
; now make my palette map
	tst.b	d3			; 8 or 24 bit BMP?
	beq	is_8			;   ... 8 bit

; A2 = pointer into sprite header
; a3 = pointer to sprite data
; a4 = pointer to mem area, PC bmp lays at space(a4)
; A5 = pointer to space for sprite
; d0 = nbr of lines -1
; d1 = nbr of bytes per line
; d2 = offset to start of last line in bitmap
; d3 = .H = alpha flag .L if -1 if 24 bitmap, 0 if 8 bitmap
; d4 = -1 if no optional params, else x,y spot
; d5 = number of pixels in sprite
; d6 = size of sprite + header in bytes
; d7 = sprite ysize | sprite xsize -1
; data in pc bitmap is BGR
is_24					; here we have a 24 bit bitmap
	lea	hlen+space(a4),a1	; point pc bmp data
	add.l	d2,a1			; point to data for last line in pc image
lp_24	move.l	a1,a0			; start here
	move.w	d7,d3
line_lp
	move.b	2(a0),d2		; R
	lsl.l	#8,d2
	move.b	1(a0),d2		; G
	lsl.l	#8,d2
	move.b	(a0),d2 		; B
	lsl.l	#8,d2
	move.l	d2,(a3)+		; put into my area
	addq.l	#3,a0
	dbf	d3,line_lp		; do it for the entire lone
	sub.l	d1,a1			; treat next line now
	dbf	d0,lp_24		; for all lines
do_mask
	swap	d3			; d3.w = 0 or pto.alph
	move.l	a3,d2
	sub.l	a2,d2
	move.l	d2,(a2)+		; pointer to mask now
	moveq	#-1,d1
	cmp.l	d1,d4			; any optional parameters (1 col to be made transparent)?
	beq.s	ful_msk 		; no, just set up opaque mask
	lea	pto_cpat(a5),a2
	add.l	(a2),a2 		; point to sprite colour data now
	move.l	a2,a1
	move.w	d4,d0			; y spot
	beq.s	no_ysp			; easy
	clr.l	d2
	move.w	d7,d2			; sprite x size -1
	addq.w	#1,d2
	lsl.w	#2,d2			; 4 bytes per pixel = line length in bytes
	mulu	d0,d2
	add.l	d2,a2			; point to line
no_ysp	swap	d4
	lsl.w	#2,d4
	move.l	(a2,d4.w),d0		; colour to look out for
	moveq	#0,d2			; transparent mask
	tst.w	d3			; aplha mask?
	bne.s	alpha			; yes (only 1 byte per pixl in mask)
mask_lp3
	cmp.l	(a1)+,d0		; it is this colour?
	bne.s	opaq_pix		;  ....no
	move.l	d2,-4(a1)		;  ... yes, set empty colour data
	move.l	d2,(a3)+		;  and set transparent mask
	subq.l	#1,d5
	bne.s	mask_lp3
	bra.s	save_it
opaq_pix
	move.l	d1,(a3)+
	subq.l	#1,d5
	bne.s	mask_lp3
	bra.s	save_it
alpha
	cmp.l	(a1)+,d0		; it is this colour?
	bne.s	opak_pix		;  ....no
	move.l	d2,-4(a1)		;  ... yes, set empty colour data
	move.b	d2,(a3)+		;  and set transparent mask
	subq.l	#1,d5
	bne.s	alpha
	bra.s	save_it
opak_pix
	move.b	d1,(a3)+
	subq.l	#1,d5
	bne.s	alpha
	bra.s	save_it
ful_msk
	tst.w	d3			; alpha mask?
	beq.s	mask_lp 		; no
	lsr.l	#2,d5			; nbr of pixels / 4 since this is an alpha mask
	addq.l	#1,d5			; if total not divisible by 4
mask_lp move.l	d1,(a3)+		; fill in mask
	subq.l	#1,d5
	bne.s	mask_lp

save_it
	moveq	#0,d0
	move.w	(a4),d0 		; any filename?
	beq.s	bmp_out2		; no
	move.l	a4,a0			; point to filename
	moveq	#1,d0
	moveq	#-1,d1
	moveq	#3,d3			; open overwrite...
	dtrap	2,bmp_out2
	move.l	d6,d2			; length of file to save
	move.l	a5,a1
	moveq	#$49,d0
	moveq	#-1,d3
	trap	#3			; greek goddess
	moveq	#2,d0
	trap	#2

bmp_out3
	move.l	a5,a0			; release ql screen mem
	subq.l	#4,a0			; rechp link
	move.l	$e0(a6),a1		; linked list
unlnk_lp
	move.l	(a1),d1 		; now find my entry
	beq	do_rel			; what, no more?
	cmp.l	a0,d1
	beq.s	do_unlink
	move.l	d1,a1
	bra.s	unlnk_lp
do_unlink
	move.l	(a0),(a1)
do_rel
	xjsr	gu_rchp
	sub.l	a5,a5
bmp_out2
	move.l	a4,a0			; release pc screen mem
	xjsr	gu_rchp
	tst.l	d0
	bne.s	out2
	move.l	a5,d2
	jmp	rt_fp			; return float
out2	rts
*----------------


; A2 = pointer into sprite header
; a3 = pointer to sprite data
; a4 = pointer to mem area, PC bmp lays at space(a4)
; A5 = pointer to space for sprite
; d0 = nbr of lines -1
; d1 = nbr of bytes per line
; d2 = offset to start of last line in bitmap
; d3 = -1 if 24 bitmap, 0 if 8 bitmap
; d4 = -1 if no optional params, else x,y spot
; d5 = number of pixels in sprite
; d6 = size of sprite + header in bytes
; d7 = sprite ysize | sprite xsize -1
; data in pc bitmap is BGR


is_8
	move.w	space+nbcol(a4),d3	; nbr of colours for this bitmap
	bne.s	no_0
	move.w	#255,d3
	bra.s	com
no_0	rol.w	#8,d3
	subq.w	#1,d3
com	lea	palet(a4),a0		; point to my future palette
	lea	space+hlen(a4),a1	; point to PC palette
	move.l	d4,-(a7)
bord32	clr.l	d4
	move.l	(a1)+,d4		; bg r0
	ror.l	#8,d4			; 0b gr
	rol.w	#8,d4			; 0b rg
	swap	d4			; rg 0b
	rol.w	#8,d4			; rg b0
	move.l	d4,(a0)+
	dbf	d3,bord32
	lea	palet(a4),a0		; point to palette
	add.l	d2,a1			; point beginning of last line
	move.l	a1,d4
bitlp1	move.l	d4,a1
	move.w	d7,d3			; nbr of pixels per line -1
bitlp2	clr.w	d2
	move.b	(a1)+,d2		; index into palette
	lsl.w	#2,d2			; into long words
	move.l	(a0,d2.w),(a3)+ 	; put into sprite data
	dbf	d3,bitlp2
	sub.l	d1,d4
	dbf	d0,bitlp1
	move.l	(sp)+,d4
	bra	do_mask

	end
